from struct import pack


class BGAPICommandPacketBuilder(object):
    @staticmethod
    def system_reset(boot_in_dfu):
        return pack('<4BB', 0, 1, 0, 0, boot_in_dfu)

    @staticmethod
    def system_hello():
        return pack('<4B', 0, 0, 0, 1)

    @staticmethod
    def system_address_get():
        return pack('<4B', 0, 0, 0, 2)

    @staticmethod
    def system_reg_write(address, value):
        return pack('<4BHB', 0, 3, 0, 3, address, value)

    @staticmethod
    def system_reg_read(address):
        return pack('<4BH', 0, 2, 0, 4, address)

    @staticmethod
    def system_get_counters():
        return pack('<4B', 0, 0, 0, 5)

    @staticmethod
    def system_get_connections():
        return pack('<4B', 0, 0, 0, 6)

    @staticmethod
    def system_read_memory(address, length):
        return pack('<4BIB', 0, 5, 0, 7, address, length)

    @staticmethod
    def system_get_info():
        return pack('<4B', 0, 0, 0, 8)

    @staticmethod
    def system_endpoint_tx(endpoint, data):
        return pack('<4BBB' + str(len(data)) + 's', 0, 2 + len(data), 0, 9,
                    endpoint, len(data), b''.join(chr(i) for i in data))

    @staticmethod
    def system_whitelist_append(address, address_type):
        return pack('<4B6sB', 0, 7, 0, 10, b''.join(chr(i) for i in address),
                    address_type)

    @staticmethod
    def system_whitelist_remove(address, address_type):
        return pack('<4B6sB', 0, 7, 0, 11, b''.join(chr(i) for i in address),
                    address_type)

    @staticmethod
    def system_whitelist_clear():
        return pack('<4B', 0, 0, 0, 12)

    @staticmethod
    def system_endpoint_rx(endpoint, size):
        return pack('<4BBB', 0, 2, 0, 13, endpoint, size)

    @staticmethod
    def system_endpoint_set_watermarks(endpoint, rx, tx):
        return pack('<4BBBB', 0, 3, 0, 14, endpoint, rx, tx)

    @staticmethod
    def flash_ps_defrag():
        return pack('<4B', 0, 0, 1, 0)

    @staticmethod
    def flash_ps_dump():
        return pack('<4B', 0, 0, 1, 1)

    @staticmethod
    def flash_ps_erase_all():
        return pack('<4B', 0, 0, 1, 2)

    @staticmethod
    def flash_ps_save(key, value):
        return pack('<4BHB' + str(len(value)) + 's', 0, 3 + len(value), 1, 3,
                    key, len(value), b''.join(chr(i) for i in value))

    @staticmethod
    def flash_ps_load(key):
        return pack('<4BH', 0, 2, 1, 4, key)

    @staticmethod
    def flash_ps_erase(key):
        return pack('<4BH', 0, 2, 1, 5, key)

    @staticmethod
    def flash_erase_page(page):
        return pack('<4BB', 0, 1, 1, 6, page)

    @staticmethod
    def flash_write_words(address, words):
        return pack('<4BHB' + str(len(words)) + 's', 0, 3 + len(words), 1, 7,
                    address, len(words), b''.join(chr(i) for i in words))

    @staticmethod
    def attributes_write(handle, offset, value):
        return pack('<4BHBB' + str(len(value)) + 's', 0, 4 + len(value), 2, 0,
                    handle, offset, len(value), b''.join(chr(i) for i in value))

    @staticmethod
    def attributes_read(handle, offset):
        return pack('<4BHH', 0, 4, 2, 1, handle, offset)

    @staticmethod
    def attributes_read_type(handle):
        return pack('<4BH', 0, 2, 2, 2, handle)

    @staticmethod
    def attributes_user_read_response(connection, att_error, value):
        return pack('<4BBBB' + str(len(value)) + 's', 0, 3 + len(value), 2, 3,
                    connection, att_error, len(value),
                    b''.join(chr(i) for i in value))

    @staticmethod
    def attributes_user_write_response(connection, att_error):
        return pack('<4BBB', 0, 2, 2, 4, connection, att_error)

    @staticmethod
    def connection_disconnect(connection):
        return pack('<4BB', 0, 1, 3, 0, connection)

    @staticmethod
    def connection_get_rssi(connection):
        return pack('<4BB', 0, 1, 3, 1, connection)

    @staticmethod
    def connection_update(connection, interval_min, interval_max,
                          latency, timeout):
        return pack('<4BBHHHH', 0, 9, 3, 2, connection, interval_min,
                    interval_max, latency, timeout)

    @staticmethod
    def connection_version_update(connection):
        return pack('<4BB', 0, 1, 3, 3, connection)

    @staticmethod
    def connection_channel_map_get(connection):
        return pack('<4BB', 0, 1, 3, 4, connection)

    @staticmethod
    def connection_channel_map_set(connection, map):
        return pack('<4BBB' + str(len(map)) + 's', 0, 2 + len(map), 3, 5,
                    connection, len(map), b''.join(chr(i) for i in map))

    @staticmethod
    def connection_features_get(connection):
        return pack('<4BB', 0, 1, 3, 6, connection)

    @staticmethod
    def connection_get_status(connection):
        return pack('<4BB', 0, 1, 3, 7, connection)

    @staticmethod
    def connection_raw_tx(connection, data):
        return pack('<4BBB' + str(len(data)) + 's', 0, 2 + len(data), 3, 8,
                    connection, len(data), b''.join(chr(i) for i in data))

    @staticmethod
    def attclient_find_by_type_value(connection, start, end, uuid, value):
        return pack('<4BBHHHB' + str(len(value)) + 's', 0, 8 + len(value), 4, 0,
                    connection, start, end, uuid, len(value),
                    b''.join(chr(i) for i in value))

    @staticmethod
    def attclient_read_by_group_type(connection, start, end, uuid):
        return pack('<4BBHHB' + str(len(uuid)) + 's', 0, 6 + len(uuid), 4, 1,
                    connection, start, end, len(uuid),
                    b''.join(chr(i) for i in uuid))

    @staticmethod
    def attclient_read_by_type(connection, start, end, uuid=[0x03, 0x28]):
        # Using the default UUID type to find custom UUIDs, which seems to make
        # querying for characteristics faster.
        return pack('<4BBHHB' + str(len(uuid)) + 's', 0, 6 + len(uuid), 4, 2,
                    connection, start, end, len(uuid),
                    b''.join(chr(i) for i in uuid))

    @staticmethod
    def attclient_find_information(connection, start, end):
        return pack('<4BBHH', 0, 5, 4, 3, connection, start, end)

    @staticmethod
    def attclient_read_by_handle(connection, chrhandle):
        return pack('<4BBH', 0, 3, 4, 4, connection, chrhandle)

    @staticmethod
    def attclient_attribute_write(connection, atthandle, data):
        return pack('<4BBHB' + str(len(data)) + 's', 0, 4 + len(data), 4, 5,
                    connection, atthandle, len(data),
                    b''.join(chr(i) for i in data))

    @staticmethod
    def attclient_write_command(connection, atthandle, data):
        return pack('<4BBHB' + str(len(data)) + 's', 0, 4 + len(data), 4, 6,
                    connection, atthandle, len(data),
                    b''.join(chr(i) for i in data))

    @staticmethod
    def attclient_indicate_confirm(connection):
        return pack('<4BB', 0, 1, 4, 7, connection)

    @staticmethod
    def attclient_read_long(connection, chrhandle):
        return pack('<4BBH', 0, 3, 4, 8, connection, chrhandle)

    @staticmethod
    def attclient_prepare_write(connection, atthandle, offset, data):
        return pack('<4BBHHB' + str(len(data)) + 's', 0, 6 + len(data), 4, 9,
                    connection, atthandle, offset, len(data),
                    b''.join(chr(i) for i in data))

    @staticmethod
    def attclient_execute_write(connection, commit):
        return pack('<4BBB', 0, 2, 4, 10, connection, commit)

    @staticmethod
    def attclient_read_multiple(connection, handles):
        return pack('<4BBB' + str(len(handles)) + 's', 0, 2 + len(handles), 4,
                    11, connection, len(handles),
                    b''.join(chr(i) for i in handles))

    @staticmethod
    def sm_encrypt_start(handle, bonding):
        return pack('<4BBB', 0, 2, 5, 0, handle, bonding)

    @staticmethod
    def sm_set_bondable_mode(bondable):
        return pack('<4BB', 0, 1, 5, 1, bondable)

    @staticmethod
    def sm_delete_bonding(handle):
        return pack('<4BB', 0, 1, 5, 2, handle)

    @staticmethod
    def sm_set_parameters(mitm, min_key_size, io_capabilities):
        return pack('<4BBBB', 0, 3, 5, 3, mitm, min_key_size, io_capabilities)

    @staticmethod
    def sm_passkey_entry(handle, passkey):
        return pack('<4BBI', 0, 5, 5, 4, handle, passkey)

    @staticmethod
    def sm_get_bonds():
        return pack('<4B', 0, 0, 5, 5)

    @staticmethod
    def sm_set_oob_data(oob):
        return pack('<4BB' + str(len(oob)) + 's', 0, 1 + len(oob), 5, 6,
                    len(oob), b''.join(chr(i) for i in oob))

    @staticmethod
    def gap_set_privacy_flags(peripheral_privacy, central_privacy):
        return pack('<4BBB', 0, 2, 6, 0, peripheral_privacy, central_privacy)

    @staticmethod
    def gap_set_mode(discover, connect):
        return pack('<4BBB', 0, 2, 6, 1, discover, connect)

    @staticmethod
    def gap_discover(mode):
        return pack('<4BB', 0, 1, 6, 2, mode)

    @staticmethod
    def gap_connect_direct(address, addr_type, conn_interval_min,
                           conn_interval_max, timeout, latency):
        return pack('<4B6sBHHHH', 0, 15, 6, 3,
                    b''.join(reversed([chr(i) for i in address])), addr_type,
                    conn_interval_min, conn_interval_max, timeout, latency)

    @staticmethod
    def gap_end_procedure():
        return pack('<4B', 0, 0, 6, 4)

    @staticmethod
    def gap_connect_selective(conn_interval_min, conn_interval_max,
                              timeout, latency):
        return pack('<4BHHHH', 0, 8, 6, 5, conn_interval_min, conn_interval_max,
                    timeout, latency)

    @staticmethod
    def gap_set_filtering(scan_policy, adv_policy, scan_duplicate_filtering):
        return pack('<4BBBB', 0, 3, 6, 6, scan_policy, adv_policy,
                    scan_duplicate_filtering)

    @staticmethod
    def gap_set_scan_parameters(scan_interval, scan_window, active):
        return pack('<4BHHB', 0, 5, 6, 7, scan_interval, scan_window, active)

    @staticmethod
    def gap_set_adv_parameters(adv_interval_min, adv_interval_max,
                               adv_channels):
        return pack('<4BHHB', 0, 5, 6, 8, adv_interval_min, adv_interval_max,
                    adv_channels)

    @staticmethod
    def gap_set_adv_data(set_scanrsp, adv_data):
        return pack('<4BBB' + str(len(adv_data)) + 's', 0, 2 + len(adv_data), 6,
                    9, set_scanrsp, len(adv_data),
                    b''.join(chr(i) for i in adv_data))

    @staticmethod
    def gap_set_directed_connectable_mode(address, addr_type):
        return pack('<4B6sB', 0, 7, 6, 10, b''.join(chr(i) for i in address),
                    addr_type)

    @staticmethod
    def hardware_io_port_config_irq(port, enable_bits, falling_edge):
        return pack('<4BBBB', 0, 3, 7, 0, port, enable_bits, falling_edge)

    @staticmethod
    def hardware_set_soft_timer(time, handle, single_shot):
        return pack('<4BIBB', 0, 6, 7, 1, time, handle, single_shot)

    @staticmethod
    def hardware_adc_read(input, decimation, reference_selection):
        return pack('<4BBBB', 0, 3, 7, 2, input, decimation,
                    reference_selection)

    @staticmethod
    def hardware_io_port_config_direction(port, direction):
        return pack('<4BBB', 0, 2, 7, 3, port, direction)

    @staticmethod
    def hardware_io_port_config_function(port, function):
        return pack('<4BBB', 0, 2, 7, 4, port, function)

    @staticmethod
    def hardware_io_port_config_pull(port, tristate_mask, pull_up):
        return pack('<4BBBB', 0, 3, 7, 5, port, tristate_mask, pull_up)

    @staticmethod
    def hardware_io_port_write(port, mask, data):
        return pack('<4BBBB', 0, 3, 7, 6, port, mask, data)

    @staticmethod
    def hardware_io_port_read(port, mask):
        return pack('<4BBB', 0, 2, 7, 7, port, mask)

    @staticmethod
    def hardware_spi_config(channel, polarity, phase, bit_order,
                            baud_e, baud_m):
        return pack('<4BBBBBBB', 0, 6, 7, 8, channel, polarity, phase,
                    bit_order, baud_e, baud_m)

    @staticmethod
    def hardware_spi_transfer(channel, data):
        return pack('<4BBB' + str(len(data)) + 's', 0, 2 + len(data), 7, 9,
                    channel, len(data), b''.join(chr(i) for i in data))

    @staticmethod
    def hardware_i2c_read(address, stop, length):
        return pack('<4BBBB', 0, 3, 7, 10, address, stop, length)

    @staticmethod
    def hardware_i2c_write(address, stop, data):
        return pack('<4BBBB' + str(len(data)) + 's', 0, 3 + len(data), 7, 11,
                    address, stop, len(data), b''.join(chr(i) for i in data))

    @staticmethod
    def hardware_set_txpower(power):
        return pack('<4BB', 0, 1, 7, 12, power)

    @staticmethod
    def hardware_timer_comparator(timer, channel, mode, comparator_value):
        return pack('<4BBBBH', 0, 5, 7, 13, timer, channel, mode,
                    comparator_value)

    @staticmethod
    def test_phy_tx(channel, length, type):
        return pack('<4BBBB', 0, 3, 8, 0, channel, length, type)

    @staticmethod
    def test_phy_rx(channel):
        return pack('<4BB', 0, 1, 8, 1, channel)

    @staticmethod
    def test_phy_end():
        return pack('<4B', 0, 0, 8, 2)

    @staticmethod
    def test_phy_reset():
        return pack('<4B', 0, 0, 8, 3)

    @staticmethod
    def test_get_channel_map():
        return pack('<4B', 0, 0, 8, 4)

    @staticmethod
    def test_debug(input):
        return pack('<4BB' + str(len(input)) + 's', 0, 1 + len(input), 8, 5,
                    len(input), b''.join(chr(i) for i in input))





class MyMeFace(object):

    id = None
    """ MyMeFace ID """

    detected_time = None
    """ detected time - datetime.time object"""

    face_image = None
    """ face image - PIL.Image object"""

    image_fetch_failed = False

    def __init__(self, id=None, detected_time=None, signature=None,
                 face_image=None, image_fetch_failed=False):
        """
        :param db: Id of the face in the SDK Faces database
        :param detected_time: Time when face was detected (datetime object)
        :param signature: Face's signature (array of 128 ints)
        :param face_image: PIL image object holding the face image
        :param image_fetch_failed: True if was unable to fetch face image after
        """
        self.id = id
        self.detected_time = detected_time
        self.signature=signature
        self.face_image = face_image
        self.image_fetch_failed = image_fetch_failed

class MyMePerson(object):
    db_id = None
    name = None
    profile_pic = None
    faces = None
    def __init__(self, db_id, name, profile_pic, faces):
        """
        :param db_id: Id of the person in the SDK People database
        :param name: Person's name
        :param profile_pic: Person's picture
        :param faces: List of faces (MyMeFace objects) belonging to the Person
        """
        self.db_id = db_id
        self.name = name
        self.profile_pic = profile_pic
        self.faces = faces

class BoundingBox(object):

    x = 0
    """ top left corner of the bounding box - x value"""

    def __init__(self, x, y, w, h):
        """
         All coordinates (x,y,w,h) are floating point numbers between 0 and 1
         which correspond to normalized locations on the full image.
         On a normalized image, the top left corner is treated as (x=0, y=0)
         and the bottom left corner is treated as (x=1, y=1)

         :param x: top left corner of the bounding box - x value
         :param y: top left corner of the bounding box - y value
         :param w: width of the bounding box
         :param h: heigth of the boudning box
         """
        self.x = x
        self.y = y
        self.w = w
        self.h = h

class OcrLine(object):
    text = None
    """text: UTF8 string containing a line of text found in the image"""
    def __init__(self, text, bounding_box):
        self.text = text
        self.bounding_box = bounding_box

class MyMeImage(object):

    def __init__(self, frame_idx, detected_time=None, full_image=None):
        """
        :param frame_idx: Frame index of image
        :param detected_time: Time when image was taken (datetime object)
        :param full_image: PIL object containing the image
        """
        self.frame_idx = frame_idx
        self.detected_time = detected_time
        self.full_image = full_image



from __future__ import division # to get correct integers division

import datetime
import struct



from myme_data_objects import MyMeFace, MyMeImage, BoundingBox, OcrLine
from utils.sdk_utils import print_and_log, ERROR, bytearray_to_image
from myme_visual_topic import *
from constants import *


HEADER_PACKET_MIN_SIE  = 1 + 4 + 4 + 4

class PacketParser(object):

    @staticmethod
    def _parse_packet_header(packet):
        """
        Packet is in the form of a bytearray
        example: bytearray(b'\xd5\xb0)' (a packet of 2 bytes)
        :param packet:
        :return:
        """
        """
        message format:
        1 byte -  header - future use
        4 bytes - total length, including header
        4 bytes - signature index (running number)
        4 bytes -  timestamp, seconds from epoch
        variable length  - signature/jpeg data
        """

        if len(packet) < HEADER_PACKET_MIN_SIE:
            return None

        # <B means little endian unsigned byte
        (header, size, id, timestamp) =  struct.unpack('<BIII', str(packet[0:13]))
        # rest of the payload, returned as is
        payload = packet[13:]

        # make sure size in header matches acutal size of packet received
        if size != len(packet):
            print_and_log('ERROR: got unexpected size of packet {0} instead of {1}. '
            'Ignoring.'.format(len(packet), size),ERROR)
            return None

        return header, size, id, timestamp, payload

    @staticmethod
    def parse_time_packet(packet):
        curr_time = struct.unpack('<I', str(packet))[0]
        return datetime.datetime.utcfromtimestamp(curr_time)

    @staticmethod
    def parse_face_image_packet(packet):
        header, size, myme_id, detected_time, raw_image_bytearray = \
            PacketParser._parse_packet_header(packet)
        if header is not None:
            return myme_id, detected_time, raw_image_bytearray
        else:
            return None, None, None

    @staticmethod
    def parse_full_image_packet(packet):
        header, size, frame_idx, timestamp, payload = \
            PacketParser._parse_packet_header(packet)
        if header is not None:
            full_image = bytearray_to_image(payload)
            return MyMeImage(frame_idx, datetime.datetime.utcfromtimestamp(timestamp),
                             full_image)
        return None


    @staticmethod
    def parse_charger_packet(packet):
        int_state = int(struct.unpack('<B', str(packet))[0])
        try:
            charger_state = ChargerState(int_state)
        except ValueError:
            print_and_log('ERROR: got unexpected charger state value {0}.'
            'Ignoring.'.format(int_state),ERROR)
            return None
        return charger_state

    @staticmethod
    def parse_version_packet(packet):
        try:
            version_str = packet.decode("utf-8")
        except Exception as ex:
            'Error trying to parse version packet.\nException Msg: {}'.format(ex)
            version_str = ''

        return version_str

    @staticmethod
    def parse_battery_state_packet(packet):
        int_state = int(struct.unpack('<B', str(packet[0]))[0])

        try:
            battery_state = BatteryState(int_state)
        except ValueError:
            print_and_log('ERROR: got unexpected battery state value {0}.'
            'Ignoring.'.format(int_state),ERROR)
            return None
        return battery_state

    @staticmethod
    def parse_battery_charge_level_packet(packet):
        """
        Battery charge level packet:
        1 byte - battery level from 0 to 100 percent (unsigned 8 bit)
        :param packet:
        :return:
        """
        try:
            int_level = int(struct.unpack('<B', str(packet))[0])

        except Exception as ex:
            print_and_log('ERROR when parsing batter charge level packet.'
                          'Packet length: {}.'.format(len(packet)))
            int_level = 0

        return int_level

    @staticmethod
    def parse_face_signature_packet(packet):

        # parse header
        header, size, myme_face_id, timestamp, payload = \
            PacketParser._parse_packet_header(packet)
        sig_num_bytes = len(payload)
        # <b means little endian signed byte
        signature_int_array = list(struct.unpack('<{}b'.format(sig_num_bytes),
                                            str(payload)))

        if header is not None:
            return myme_face_id, timestamp, signature_int_array, header
        return None


    @staticmethod
    def parse_imagenet_packet(packet):
        header, size, frame_idx, timestamp, payload = PacketParser._parse_packet_header(packet)

        if header is None:
            return None

        visual_topcis_list = []
        i = 0
        while i < len(payload):
            # <B means little endian unsigned byte
            probability = struct.unpack('<B', str(payload[i:i+1]))[0]
            i += 1

            j = i # store word start index

            # find end of the (null terminated string).
            # also stop if reached the end of the payload
            # (should not happen in well formatted packet)
            while payload[i] != 0 and i < len(payload):
                i += 1

            # decode to string
            topic = payload[j:i].decode("utf-8")

            # create the current visual_topic
            visual_topcis_list.append(VisualTopic(topic, probability))

            # continue to next topic
            i += 1

        return visual_topcis_list, datetime.datetime.utcfromtimestamp(timestamp)

    # takes an integer input which is expceted to be between 0-65535
    # normilizes the value between 0-1 returning the (float) result
    @staticmethod
    def bounding_box_coordinate_to_float(x):

        MAX_UNSIGNEDD_SHORT = 65535
        if x < 0 or x > MAX_UNSIGNEDD_SHORT:
            print_and_log('ERROR: got unexpected OCR '
                          'bounding box coordinate {0}.'
                          .format(x),ERROR)
            return 0
        return x/65535

    @staticmethod
    def parse_barcode_packet(packet):
        # parse header
        flag, size, frame_idx, timestamp, payload = PacketParser._parse_packet_header(packet)
        i = 0

        # find end (null terminated string).
        # also stop if reached the end of the payload
        # (should not happen in well formatted packet)
        while payload[i] != 0 and i < len(payload):
            i += 1

        # decode to string, default of ascii is used
        barcode = payload[0:i].decode()

        return barcode, datetime.datetime.utcfromtimestamp(timestamp)

    @staticmethod
    def parse_ocr_packet(packet):
        # parse header
        # flag: 0: normal ocr line packet
        #       1: "OCR-Frame-end" packet (packet may be empty or contain OCR line(s))
        #
        flag, size, frame_idx, timestamp, payload = PacketParser._parse_packet_header(packet)

        # now data itself
        # 1 byte - future use
        # bounding box, given using shorts (2 bytes)
        # 2 bytes - x pos
        # 2 bytes - y pos
        # 2 bytes - width
        # 2 bytes - height
        # 4 bytes - length of text line below,
        #  			 not including the length field itself, including null termination
        #  <variable length> - null terminated utf8 encoded string
        i = 0
        ocr_lines_list = []
        while i < len(payload):
            i += 1 # skip the first 'future-use' byte

            # prase boundinb box
            (x, y ,w ,h) = struct.unpack('<HHHH', str(payload[i:i+8]))

            box = BoundingBox(PacketParser.bounding_box_coordinate_to_float(x),
                            PacketParser.bounding_box_coordinate_to_float(y),
                            PacketParser.bounding_box_coordinate_to_float(w),
                            PacketParser.bounding_box_coordinate_to_float(h))

            # string length
            i += 8
            str_len = struct.unpack('<I', str(payload[i:i+4]))[0]

            i+=4
            j = i # store string start index

            # find end of the (null terminated string).
            # also stop if reached the end of the payload
            # (should not happen in well formatted packet)
            while payload[i] != 0 and i < len(payload):
                i += 1

            # validate length of the string matches the one speicifed in the packet
            if str_len != i-j:
                print_and_log('got unexpected ocr line length '
                          '{0} instead of {1}.'
                          .format(str_len, i-j),ERROR)

            # decode to string
            try:
                text_line = payload[j:i].decode("utf-8")
                # append to list
                ocr_lines_list.append(OcrLine(text_line, box))
            except UnicodeDecodeError:
                print_and_log('error decoding string {0}.'
                        .format(payload[j:i]),ERROR)

            # continue to next line
            i += 1

        return ocr_lines_list, datetime.datetime.utcfromtimestamp(timestamp), flag






"""
Exceptions for pygatt Module.
"""


class BLEError(Exception):
    """Exception class for pygatt."""
    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.message)

    def __str__(self):
        return repr(self)


class NotConnectedError(BLEError):
    pass


class NotificationTimeout(BLEError):
    pass


class NoResponseError(BLEError):
    pass




import abc
from constants import *

class MyMeDelegates():
    """
    This class provides an interface for the delegate functions that the SDK
    user may choose to implement. Any function that the user does not implement
    will have the default implementation defined here
    """
    __metaclass__ = abc.ABCMeta
    def __init__(self):
        pass

    def ConnectedToDevice(self, device_address, error):
        """
        This deleagte is invoked after an attempt is made to connect to a MyMe
        device. On success, device_addr and device_name are the address/name
        of the MyMe device that user connected to. On failure, error contains
        the error message.
        :param device_addr: address of the device that user connected to
        :param error: on connection failure, contains error string
        """
        pass

    def ScanComplete(self, device_list, scan_result, error=None):
        """
        Called when a scan (following a call to scan_for_devices_with_timeout)
        is completed.
        :param device_list: List of devices.
        Each device is represented by dict with the following keys:
            "address" (device's MAC address)
            "name" (device name)
            "rssi" (signal strength)
        :param scan_result: scan compeltion reason (ScanState)
        :param error: in case of error, contains error message
        """
        pass

    def PairedDeviceStateUpdated(self, new_device_state):
        """
        This delegate is invoked when the MyMe device connection state changes
        :param state - The current device state
        """
        pass
    
    def FaceDetected(self, face, person, probability, is_last_face_in_frame):
        """
        This delegate is invoked when the MyMe device detects a face
        :param face: MyMeFace object holding the matched face does not include
                     the actual image nor image signature.
        :param person: MyMePerson object corresponding to the matched person
        :param probability: probability that <face> belongs to <person>
        :param is_last_face_in_frame: whether this is the last face for this franme
        """
        pass

    def NewPersonCreated(self, person, timestamp):
        """
        This delegate is invoked when the MyMe device detects multiple faces
        and recognizes all of them as the same person. The DB creates a new
        person with the faces.
        :param person: newly created person (MyMePerson object)
        :param timestamp: time when face was created (seconds from epoch float)
        """
        pass

    def VisualTopicsReceived(self, topics, timestamp):
        """
        This delegate is invoked when a new set of words is sent from the device.
        :param topics: list containing objects of type VisualTopic which
        contains topic and its probability of being in the frame. The list is
        sorted by probability.
        :param timestamp: the time that the words were sent from the device
        """
        pass

    def OcrLineFound(self, myme_ocr_line, timestamp):
        """
        Inovked when a line of text is detected in an image.
        :param myme_ocr_line: Line of text detected (OcrLine object).
                              None is sent in case this is the last line at this time
                              Note None might be sent even if no lines were sent, in case no
                              lines were deteceted at this time
        :param timestamp: Timestamp when line was detected
        """
        pass

    def BarcodeDetected(self, barcode, timestamp):
        """
        Invoked when a product barcode is detected in an image.
        :param barcode: The detected barcode (ascii string)
        :param timestamp: Time when barcode was detected
        """
        pass

    def ChargerStateUpdated(self, new_charger_state):
        """
        This delegate is invoked when the charger state is changed.
        :param new_charger_state: the state of the charger after the change
        """
        pass

    def DeviceTapped(self):
        """
        This delegate is called when the device is tapped
        """
        pass

    def BatteryStateUpdated(self, battery_state):
        """
        Called when the battery state is updated
        :param battery_state: BatteryState enum with the following options:
                            BatteryFull = 0
                            BatteryOK = 1
                            BatteryLow = 2
                            BatteryCritical = 3
        """
        pass


    def BatteryChargeLevelChanged(self, battery_level):
        """
        Called when the battery level is changed (expected to be called about
        once a minute while the device is connected)
        :param battery_level: Int from 1 to 100 (100 means charged 100%)
        """
        pass


    def QrCodeDetected(self, qr_code, timestamp):
        raise NotImplementedError
